from abc import ABC, abstractmethod
import logging
import pathlib
import re
from typing import Dict, List, Optional, Tuple
import uuid

import requests

from anyscale.authenticate import AuthenticationBlock, get_auth_api_client
from anyscale.client.openapi_client.api.default_api import DefaultApi as InternalApi
from anyscale.client.openapi_client.models import (
    ArchiveStatus,
    CloudDataBucketFileType,
    CloudDataBucketPresignedUploadInfo,
    CloudDataBucketPresignedUploadRequest,
    ComputeTemplateConfig,
    ComputeTemplateQuery,
    CreateComputeTemplate,
    DecoratedComputeTemplate,
)
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as ExternalApi
from anyscale.sdk.anyscale_client.configuration import Configuration
from anyscale.sdk.anyscale_client.models import (
    ApplyServiceModel,
    Cloud,
    Cluster,
    ClusterCompute,
    ClusterComputeConfig,
    ClusterEnvironment,
    ClusterEnvironmentBuild,
    ClusterEnvironmentBuildStatus,
    Project,
    RollbackServiceModel,
    ServiceEventCurrentState,
    ServiceModel,
)
from anyscale.sdk.anyscale_client.models.create_cluster_environment import (
    CreateClusterEnvironment,
)
from anyscale.sdk.anyscale_client.models.create_cluster_environment_build import (
    CreateClusterEnvironmentBuild,
)
from anyscale.sdk.anyscale_client.rest import ApiException
from anyscale.util import (
    get_cluster_model_for_current_workspace,
    get_endpoint,
    is_anyscale_workspace,
)
from anyscale.utils.runtime_env import (
    is_workspace_dependency_tracking_disabled,
    WORKSPACE_REQUIREMENTS_FILE_PATH,
    zip_local_dir,
)
from anyscale.utils.workspace_notification import (
    WORKSPACE_NOTIFICATION_ADDRESS,
    WorkspaceNotification,
)


logger = logging.getLogger(__name__)

# TODO(edoakes): figure out a sane policy for this.
# Maybe just make it part of the release process to update it, or fetch the
# default builds and get the latest one. The best thing to do is probably
# to populate this in the backend.
DEFAULT_RAY_VERSION = "2.9.2"
DEFAULT_PYTHON_VERSION = "py310"
RUNTIME_ENV_PACKAGE_FORMAT = "pkg_{content_hash}.zip"

# All workspace cluster names should start with this prefix.
WORKSPACE_CLUSTER_NAME_PREFIX = "workspace-cluster-"

OPENAPI_NO_VALIDATION = Configuration()
OPENAPI_NO_VALIDATION.client_side_validation = False


class AnyscaleClientWrapper(ABC):
    @abstractmethod
    def get_service_ui_url(self, service_id: str) -> str:
        """Get a URL to the webpage for a service."""
        raise NotImplementedError

    @abstractmethod
    def inside_workspace(self) -> bool:
        """Returns true if this code is running inside a workspace."""
        raise NotImplementedError

    @abstractmethod
    def get_workspace_requirements_path(self) -> Optional[str]:
        """Returns the path to the workspace-managed requirements file.

        Returns None if dependency tracking is disable or the file does not exist.
        """
        raise NotImplementedError

    @abstractmethod
    def get_current_workspace_cluster(self) -> Optional[Cluster]:
        """Get the cluster model for the workspace this code is running in.

        Returns None if not running in a workspace.
        """
        raise NotImplementedError

    def send_workspace_notification(self, notification: WorkspaceNotification):
        """Send a workspace notification to be displayed to the user.

        This is a no-op if called from outside a workspace.
        """
        raise NotImplementedError

    @abstractmethod
    def get_project_id(self) -> str:
        """Get the default project ID for this user.

        If running in a workspace, returns the workspace project ID.
        """
        raise NotImplementedError

    @abstractmethod
    def get_cloud_id(
        self, *, cloud_name: Optional[str] = None, compute_config_id: Optional[str]
    ) -> Optional[str]:
        """Get the cloud ID for the provided cloud name or compute config ID.

        If both arguments are None:
            - if running in a workspace, get the workspace's cloud ID.
            - else, get the user's default cloud ID.
        """
        raise NotImplementedError

    @abstractmethod
    def create_anonymous_compute_config(self, config: ComputeTemplateConfig) -> str:
        """Create an anonymous compute config and return its ID."""
        raise NotImplementedError

    @abstractmethod
    def get_compute_config_id(
        self, compute_config_name: Optional[str] = None
    ) -> Optional[str]:
        """Get the compute config ID for the provided name.

        If compute_config_name is None:
            - if running in a workspace, get the workspace's compute config ID.
            - else, get the user's default compute config ID.

        Returns None if the compute config name does not exist.
        """
        raise NotImplementedError

    @abstractmethod
    def get_compute_config(
        self, compute_config_id: str
    ) -> Optional[DecoratedComputeTemplate]:
        """Get the compute config for the provided ID.

        Returns None if not found.
        """
        raise NotImplementedError

    @abstractmethod
    def get_default_build_id(self) -> str:
        """Get the default build id.

        If running in a workspace, it will return the workspace's cluster environment build ID.
        Else it will return the default cluster environment build ID.
        """
        raise NotImplementedError

    @abstractmethod
    def get_cluster_env_build_id(
        self, image_uri: str,
    ):
        """Get the cluster env build ID for the provided name.

        If cluster_env_name is None:
            - if running in a workspace, get the workspace's cluster env build ID.
            - else, get the user's default cluster env build ID.

        Returns None if the cluster env name does not exist.
        """
        raise NotImplementedError

    @abstractmethod
    def get_cluster_env_name(self, cluster_env_build_id: str) -> Optional[str]:
        """Get the cluster env name for the provided build ID.

        The returned name will include the revision, e.g., "env-name:1".

        Returns None if not found.
        """
        raise NotImplementedError

    @abstractmethod
    def get_service(self, name: str) -> Optional[ServiceModel]:
        """Get a service by name.

        Returns None if not found.
        """
        raise NotImplementedError

    @abstractmethod
    def rollout_service(self, model: ApplyServiceModel) -> ServiceModel:
        """Deploy or update the service to use the provided config.

        Returns the service ID.
        """
        raise NotImplementedError

    @abstractmethod
    def rollback_service(
        self, service_id: str, *, max_surge_percent: Optional[int] = None
    ):
        """Roll the service back to the primary version.

        This can only be used during an active rollout.
        """
        raise NotImplementedError

    @abstractmethod
    def terminate_service(self, service_id: str):
        """Mark the service to be terminated asynchronously."""
        raise NotImplementedError

    @abstractmethod
    def upload_local_dir_to_cloud_storage(
        self, local_dir: str, *, cloud_id: str, excludes: Optional[List[str]] = None
    ) -> str:
        """Upload the provided directory to cloud storage and return a URI for it.

        The directory will be zipped and the resulting URI can be in a Ray runtime_env.

        The upload is preformed using a pre-signed URL fetched from Anyscale, so no
        local cloud provider authentication is required.
        """
        raise NotImplementedError


class FakeAnyscaleClient(AnyscaleClientWrapper):
    BASE_UI_URL = "http://fake.com"
    CLOUD_BUCKET = "s3://fake-bucket/{cloud_id}"
    DEFAULT_CLOUD_ID = "fake-default-cloud-id"
    DEFAULT_PROJECT_ID = "fake-default-project-id"
    DEFAULT_CLUSTER_COMPUTE_ID = "fake-default-cluster-compute-id"
    DEFAULT_CLUSTER_ENV_BUILD_ID = "fake-default-cluster-env-build-id"

    WORKSPACE_CLOUD_ID = "fake-workspace-cloud-id"
    WORKSPACE_CLUSTER_ID = "fake-workspace-cluster-id"
    WORKSPACE_PROJECT_ID = "fake-workspace-project-id"
    WORKSPACE_CLUSTER_COMPUTE_ID = "fake-workspace-cluster-compute-id"
    WORKSPACE_CLUSTER_ENV_BUILD_ID = "fake-workspace-cluster-env-build-id"

    def get_service_ui_url(self, service_id: str) -> str:
        return f"{self.BASE_UI_URL}/services/{service_id}"

    def __init__(self):
        self._image_uri_to_build_id: Dict[str, str] = {}
        self._build_id_to_image_uri: Dict[str, str] = {}
        self._compute_config_name_to_id: Dict[str, str] = {}
        self._compute_config_id_to_cloud_id: Dict[str, str] = {}
        self._compute_configs: Dict[str, ClusterCompute] = {}
        self._workspace_cluster: Optional[Cluster] = None
        self._workspace_dependency_tracking_enabled: bool = False
        self._services: Dict[str, ServiceModel] = {}
        self._rolled_out_model: Optional[ApplyServiceModel] = None
        self._sent_workspace_notifications: List[WorkspaceNotification] = []
        self._rolled_back_service: Optional[Tuple[str, Optional[int]]] = None
        self._terminated_service: Optional[str] = None

    def set_inside_workspace(
        self,
        inside_workspace: bool,
        *,
        requirements_path: Optional[str] = None,
        cluster_name: Optional[str] = None,
    ):
        self._requirements_path = requirements_path
        if inside_workspace:
            self._workspace_cluster = Cluster(
                id=self.WORKSPACE_CLUSTER_ID,
                name=cluster_name
                if cluster_name is not None
                else WORKSPACE_CLUSTER_NAME_PREFIX + "test",
                project_id=self.WORKSPACE_PROJECT_ID,
                cluster_compute_id=self.WORKSPACE_CLUSTER_COMPUTE_ID,
                cluster_environment_build_id=self.WORKSPACE_CLUSTER_ENV_BUILD_ID,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        else:
            self._workspace_cluster = None

    def inside_workspace(self) -> bool:
        return self.get_current_workspace_cluster() is not None

    def get_workspace_requirements_path(self) -> Optional[str]:
        return self._requirements_path

    def get_current_workspace_cluster(self) -> Optional[Cluster]:
        return self._workspace_cluster

    @property
    def sent_workspace_notifications(self) -> List[WorkspaceNotification]:
        return self._sent_workspace_notifications

    def send_workspace_notification(self, notification: WorkspaceNotification):
        if self.inside_workspace():
            self._sent_workspace_notifications.append(notification)

    def get_project_id(self) -> str:
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.project_id

        return self.DEFAULT_PROJECT_ID

    def set_compute_config_id_to_cloud_id_mapping(
        self, compute_config_id: str, cloud_id: str
    ):
        self._compute_config_id_to_cloud_id[compute_config_id] = cloud_id

    def get_cloud_id(
        self,
        *,
        cloud_name: Optional[str] = None,
        compute_config_id: Optional[str] = None,
    ) -> Optional[str]:
        assert not (cloud_name and compute_config_id)
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return self.WORKSPACE_CLOUD_ID

        if compute_config_id is not None:
            return self._compute_config_id_to_cloud_id.get(compute_config_id)

        if cloud_name is None:
            return self.DEFAULT_CLOUD_ID

        # TODO(edoakes): uncomment the below once non-default cloud is supported.
        # return f"{cloud_name}-fake-id"
        raise NotImplementedError("Only default cloud is currently implemented.")

    def add_compute_config(self, compute_config: DecoratedComputeTemplate):
        self._compute_configs[compute_config.id] = compute_config
        self._compute_config_name_to_id[compute_config.name] = compute_config.id

    def create_anonymous_compute_config(self, config: ComputeTemplateConfig) -> str:
        compute_config_id = f"compute-config-id-{str(uuid.uuid4())}"
        self.add_compute_config(
            DecoratedComputeTemplate(
                id=compute_config_id,
                name=f"compute-config-{str(uuid.uuid4())}",
                config=config,
                anonymous=True,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )
        return compute_config_id

    def get_compute_config(
        self, compute_config_id: str
    ) -> Optional[DecoratedComputeTemplate]:
        return self._compute_configs.get(compute_config_id, None)

    def get_compute_config_id(
        self, compute_config_name: Optional[str] = None
    ) -> Optional[str]:
        if compute_config_name is not None:
            return self._compute_config_name_to_id.get(compute_config_name)

        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.cluster_compute_id

        return self.DEFAULT_CLUSTER_COMPUTE_ID

    def set_image_uri_mapping(self, image_uri: str, build_id: str):
        self._image_uri_to_build_id[image_uri] = build_id
        self._build_id_to_image_uri[build_id] = image_uri

    def get_default_build_id(self) -> str:
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.cluster_environment_build_id

        return self.DEFAULT_CLUSTER_ENV_BUILD_ID

    def get_cluster_env_build_id(self, image_uri: str) -> str:
        if image_uri in self._image_uri_to_build_id:
            return self._image_uri_to_build_id[image_uri]
        else:
            raise ValueError(f"The image_uri '{image_uri}' does not exist.")

    def get_cluster_env_name(self, cluster_env_build_id: str) -> Optional[str]:
        return self._build_id_to_image_uri.get(cluster_env_build_id, None)

    def update_service(self, model: ServiceModel):
        self._services[model.id] = model

    def get_service(self, name: str) -> Optional[ServiceModel]:
        for service in self._services.values():
            if service.name == name:
                return service

        return None

    @property
    def rolled_out_model(self) -> Optional[ApplyServiceModel]:
        return self._rolled_out_model

    def rollout_service(self, model: ApplyServiceModel) -> ServiceModel:
        self._rolled_out_model = model
        existing_service = self.get_service(model.name)
        if existing_service is not None:
            service_id = existing_service.id
        else:
            service_id = f"service-id-{str(uuid.uuid4())}"

        service = ServiceModel(
            id=service_id,
            name=model.name,
            current_state=ServiceEventCurrentState.RUNNING,
            base_url="http://fake-service-url",
            auth_token="fake-auth-token",
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )

        self.update_service(service)
        return service

    @property
    def rolled_back_service(self) -> Optional[Tuple[str, Optional[int]]]:
        return self._rolled_back_service

    def rollback_service(
        self, service_id: str, *, max_surge_percent: Optional[int] = None
    ):
        self._rolled_back_service = (service_id, max_surge_percent)

    @property
    def terminated_service(self) -> Optional[str]:
        return self._terminated_service

    def terminate_service(self, service_id: str):
        self._terminated_service = service_id

    def upload_local_dir_to_cloud_storage(
        self,
        local_dir: str,  # noqa: ARG002
        *,
        cloud_id: str,
        excludes: Optional[List[str]] = None,  # noqa: ARG002
    ) -> str:
        bucket = self.CLOUD_BUCKET.format(cloud_id=cloud_id)
        return f"{bucket}/fake_pkg_{uuid.uuid4()}.zip"


class RealAnyscaleClient(AnyscaleClientWrapper):
    # Number of entries to fetch per request for list endpoints.
    LIST_ENDPOINT_COUNT = 50

    def __init__(
        self,
        *,
        api_clients: Optional[Tuple[ExternalApi, InternalApi]] = None,
        workspace_requirements_file_path: str = WORKSPACE_REQUIREMENTS_FILE_PATH,
    ):
        if api_clients is None:
            auth_block: AuthenticationBlock = get_auth_api_client(
                raise_structured_exception=True
            )
            api_clients = (auth_block.anyscale_api_client, auth_block.api_client)

        self._external_api_client, self._internal_api_client = api_clients
        self._workspace_requirements_file_path = workspace_requirements_file_path

        # Cached IDs and models to avoid duplicate lookups.
        self._default_project_id: Optional[str] = None
        self._cloud_id_cache: Dict[Optional[str], str] = {}
        self._current_workspace_cluster: Optional[Cluster] = None

    def get_service_ui_url(self, service_id: str) -> str:
        return get_endpoint(f"/services/{service_id}")

    def inside_workspace(self) -> bool:
        return self.get_current_workspace_cluster() is not None

    def get_workspace_requirements_path(self) -> Optional[str]:
        if (
            not self.inside_workspace()
            or is_workspace_dependency_tracking_disabled()
            or not pathlib.Path(self._workspace_requirements_file_path).is_file()
        ):
            return None

        return self._workspace_requirements_file_path

    def get_current_workspace_cluster(self) -> Optional[Cluster]:
        # Checks for the existence of the ANYSCALE_EXPERIMENTAL_WORKSPACE_ID env var.
        if not is_anyscale_workspace():
            return None

        if self._current_workspace_cluster is None:
            # Picks up the cluster ID from the ANYSCALE_SESSION_ID env var.
            self._current_workspace_cluster = get_cluster_model_for_current_workspace(
                self._external_api_client
            )

        return self._current_workspace_cluster

    def get_project_id(self) -> str:
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.project_id

        if self._default_project_id is None:
            default_project: Project = self._external_api_client.get_default_project().result
            self._default_project_id = default_project.id

        return self._default_project_id

    def _get_cloud_id_for_compute_config_id(self, compute_config_id: str) -> str:
        cluster_compute: ClusterCompute = self._external_api_client.get_cluster_compute(
            compute_config_id
        ).result
        cluster_compute_config: ClusterComputeConfig = cluster_compute.config
        return cluster_compute_config.cloud_id

    def get_cloud_id(
        self, cloud_name: Optional[str] = None, compute_config_id: Optional[str] = None
    ) -> Optional[str]:
        if cloud_name is not None:
            raise NotImplementedError("Only default cloud is currently implemented.")

        if compute_config_id is not None:
            return self._get_cloud_id_for_compute_config_id(compute_config_id)

        if cloud_name in self._cloud_id_cache:
            return self._cloud_id_cache[cloud_name]

        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            # NOTE(edoakes): the Cluster model has a compute_config_config model that includes
            # its cloud ID, but it's not always populated.
            # TODO(edoakes): add cloud_id to the Cluster model to avoid a second RTT.
            cloud_id = self._get_cloud_id_for_compute_config_id(
                workspace_cluster.cluster_compute_id
            )
        else:
            result: Cloud = self._external_api_client.get_default_cloud().result
            cloud_id = result.id

        self._cloud_id_cache[cloud_name] = cloud_id
        return cloud_id

    def create_anonymous_compute_config(self, config: ComputeTemplateConfig) -> str:
        return self._internal_api_client.create_compute_template_api_v2_compute_templates_post(
            create_compute_template=CreateComputeTemplate(
                config=config, anonymous=True,
            )
        ).result.id

    def get_compute_config(
        self, compute_config_id: str
    ) -> Optional[DecoratedComputeTemplate]:
        try:
            cluster_compute: ClusterCompute = self._internal_api_client.get_compute_template_api_v2_compute_templates_template_id_get(
                compute_config_id
            ).result
            return cluster_compute
        except ApiException as e:
            if e.status == 404:
                return None

            raise e from None

    def get_compute_config_id(
        self, compute_config_name: Optional[str] = None
    ) -> Optional[str]:
        if compute_config_name is not None:
            cluster_computes = self._internal_api_client.search_compute_templates_api_v2_compute_templates_search_post(
                ComputeTemplateQuery(
                    orgwide=True,
                    name={"equals": compute_config_name},
                    include_anonymous=True,
                    archive_status=ArchiveStatus.ALL,
                    # TODO(edoakes): implement version strings.
                    version=None,
                )
            ).results

            if len(cluster_computes) == 0:
                return None

            compute_template: DecoratedComputeTemplate = cluster_computes[0]
            return compute_template.id

        # If the compute config name is not provided, we pick an appropriate default.
        #
        #   - If running in a workspace, we use the workspace's compute config, with
        #     a minor transformation if auto_select_worker_config is applied.
        #
        #   - Otherwise, we use the default compute config provided by the API.

        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            workspace_compute_config: DecoratedComputeTemplate = self.get_compute_config(
                workspace_cluster.cluster_compute_id
            )
            workspace_config: ClusterComputeConfig = workspace_compute_config.config
            if workspace_config.auto_select_worker_config:
                workspace_config = self._apply_standardized_head_node_type(
                    workspace_config
                )
                return self.create_anonymous_compute_config(workspace_config)
            else:
                return workspace_cluster.cluster_compute_id

        # NOTE(edoakes): for some reason the API endpoint is called "cluster compute" but this
        # is the non-deprecated API to use for compute configs.
        return self._external_api_client.get_default_cluster_compute().result.id

    def _apply_standardized_head_node_type(
        self, compute_config: ClusterComputeConfig
    ) -> ClusterComputeConfig:
        """
        Apply the following transformations to the provided compute config:

        1. Standardize the head node instance type.
        2. Disable scheduling on the head node.
        """
        # Retrieve the default cluster compute config.
        default_compute_config: DecoratedComputeTemplate = self._external_api_client.get_default_cluster_compute().result.config

        # Standardize the head node instance type.
        compute_config.head_node_type.instance_type = (
            default_compute_config.head_node_type.instance_type
        )

        # Disable scheduling on the head node.
        if compute_config.head_node_type.resources is None:
            compute_config.head_node_type.resources = {}
        compute_config.head_node_type.resources["CPU"] = 0

        return compute_config

    def get_cluster_env_name(self, cluster_env_build_id: str) -> Optional[str]:
        try:
            build: ClusterEnvironmentBuild = self._external_api_client.get_cluster_environment_build(
                cluster_env_build_id
            ).result
            cluster_env: ClusterEnvironment = self._external_api_client.get_cluster_environment(
                build.cluster_environment_id
            ).result
            return f"{cluster_env.name}:{build.revision}"
        except ApiException as e:
            if e.status == 404:
                return None

            raise e from None

    def get_default_build_id(self) -> str:
        """Get default build id.

        If running in a workspace, it will return the workspace's cluster environment build ID.
        Else it will return the default cluster environment build ID.
        """
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.cluster_environment_build_id
        result: ClusterEnvironmentBuild = self._external_api_client.get_default_cluster_environment_build(
            DEFAULT_PYTHON_VERSION, DEFAULT_RAY_VERSION,
        ).result
        return result.id

    @staticmethod
    def _get_cluster_env_name_from_image_uri(image_uri: str) -> str:
        if len(image_uri) == 0:
            raise ValueError("image_uri cannot be empty.")

        pattern = re.compile("^[A-Za-z0-9_-]+$")
        # Keep only characters that match the pattern
        escaped = []
        for c in image_uri:
            if not pattern.match(c):
                escaped.append("-")
            else:
                escaped.append(c)
        return "".join(escaped)

    def _get_cluster_env_by_name(self, name: str) -> Optional[ClusterEnvironment]:
        cluster_envs = self._external_api_client.search_cluster_environments(
            {
                "name": {"equals": name},
                "paging": {"count": 1},
                "include_anonymous": True,
            }
        ).results
        return cluster_envs[0] if cluster_envs else None

    def get_cluster_env_build_id(self, image_uri: str) -> str:
        """Get the cluster environment build ID for the cluster environment with provided image_uri.

        It maps a image_uri to a cluster env name and then use the name to get the cluster env.
        If there exists a cluster environment for the image_uri, it will reuse the cluster env.
        Else it will create a new cluster environment.
        It will create a new cluster environment build with the provided image_uri or try to reuse one with the same image_uri if exists.

        The same URI should map to the same cluster env name and therefore the build but it is not guaranteed since
        the name format can change.

        """
        cluster_env_name = self._get_cluster_env_name_from_image_uri(image_uri)

        existing_cluster_env = self._get_cluster_env_by_name(cluster_env_name)

        if existing_cluster_env is None:
            # this creates a cluster env only and it does not trigger a build to avoid the race condition
            # when creating a cluster environment and a build at the same time and then list builds for the cluster environment later.
            # ```
            #     cluster_env == self._external_api_client.create_cluster_environment(..., image_uri=image_uri)
            #     build == self._external_api_client.create_cluster_environment_build(..., cluster_env_id=cluster_env.id)
            # ```
            # The race condition can happen if another build for the same cluster envrionment is created between the two calls.

            cluster_environment = self._external_api_client.create_cluster_environment(
                CreateClusterEnvironment(name=cluster_env_name, anonymous=True)
            ).result
            cluster_env_id = cluster_environment.id
        else:
            cluster_env_id = existing_cluster_env.id
            # since we encode the image_uri into the cluster env name, there should exist one and only one build that matches the image_uri.
            cluster_env_builds = self._external_api_client.list_cluster_environment_builds(
                cluster_environment_id=cluster_env_id, count=1
            ).results
            build = cluster_env_builds[0] if cluster_env_builds else None
            if (
                build is not None
                and build.docker_image_name == image_uri
                and build.status == ClusterEnvironmentBuildStatus.SUCCEEDED
            ):
                return build.id

        # Still create a new build if the cluster env already exists but the build does not match the image_uri.
        result = self._external_api_client.create_cluster_environment_build(
            CreateClusterEnvironmentBuild(
                # For historical reasons, we have to use docker_image_name instead of image_uri; but it is just a URI to the image.
                cluster_environment_id=cluster_env_id,
                docker_image_name=image_uri,
            )
        ).result

        assert result.completed
        return result.cluster_environment_build_id

    def send_workspace_notification(
        self, notification: WorkspaceNotification,
    ):
        if not self.inside_workspace():
            return

        try:
            r = requests.post(WORKSPACE_NOTIFICATION_ADDRESS, json=notification.dict())
            r.raise_for_status()
        except Exception:
            logger.exception(
                "Failed to send workspace notification. "
                "This should not happen, so please contact Anyscale support."
            )

    def get_service(self, name: str) -> Optional[ServiceModel]:
        # TODO(edoakes): this endpoint is very slow and there's no reason we should need
        # to use this complex list endpoint just to fetch a service by name.
        paging_token = None
        project_id = self.get_project_id()
        service: Optional[ServiceModel] = None
        while True:
            resp = self._external_api_client.list_services(
                project_id=project_id,
                name=name,
                count=self.LIST_ENDPOINT_COUNT,
                paging_token=paging_token,
            )
            for result in resp.results:
                if result.name == name:
                    service = result
                    break

            paging_token = resp.metadata.next_paging_token
            if service is not None or paging_token is None:
                break

        return service

    def rollout_service(self, model: ApplyServiceModel) -> ServiceModel:
        result: ServiceModel = self._external_api_client.rollout_service(model).result
        return result

    def rollback_service(
        self, service_id: str, *, max_surge_percent: Optional[int] = None
    ) -> ServiceModel:
        result: ServiceModel = self._external_api_client.rollback_service(
            service_id,
            rollback_service_model=RollbackServiceModel(
                max_surge_percent=max_surge_percent
            ),
        )
        return result

    def terminate_service(self, service_id: str) -> ServiceModel:
        result: ServiceModel = self._external_api_client.terminate_service(service_id)
        return result

    def upload_local_dir_to_cloud_storage(
        self, local_dir: str, *, cloud_id: str, excludes: Optional[List[str]] = None
    ) -> str:
        if not pathlib.Path(local_dir).is_dir():
            raise RuntimeError(f"Path '{local_dir}' is not a valid directory.")

        with zip_local_dir(local_dir, excludes=excludes) as (
            _,
            zip_file_bytes,
            content_hash,
        ):
            request = CloudDataBucketPresignedUploadRequest(
                file_type=CloudDataBucketFileType.RUNTIME_ENV_PACKAGES,
                file_name=RUNTIME_ENV_PACKAGE_FORMAT.format(content_hash=content_hash),
            )
            info: CloudDataBucketPresignedUploadInfo = self._internal_api_client.generate_cloud_data_bucket_presigned_upload_url_api_v2_clouds_cloud_id_generate_cloud_data_bucket_presigned_upload_url_post(
                cloud_id, request
            ).result
            requests.put(info.upload_url, data=zip_file_bytes).raise_for_status()

        return info.file_uri



import copy
from enum import Enum
import re
from typing import Any, Dict, List, Optional, Union

import yaml


IMAGE_URI_PATTERN = re.compile(
    r"^"
    # Optional registry host: hostname with optional port
    r"((?P<host>[a-zA-Z0-9.-]+)(?::(?P<port>[0-9]+))?/)?"
    # Repository name: user_name/repository
    r"(?P<repository>[a-zA-Z0-9-_]+/([a-zA-Z0-9-/_])+)"
    # Optional Tag: version or string after ':'
    r"(:(?P<tag>[a-zA-Z0-9_.-]+))?"
    # Optional Digest: string after '@'
    # Note that when both tag and digest are provided, the tag is ignored.
    r"(@(?P<digest>[a-zA-Z0-9]+))?"
    r"$"
)


class ServiceConfig:
    def __init__(
        self,
        *,
        applications: List[Dict[str, Any]],
        name: Optional[str] = None,
        image_uri: Optional[str] = None,
        compute_config: Optional[Union[Dict, str]] = None,
        working_dir: Optional[str] = None,
        excludes: Optional[List[str]] = None,
        requirements: Optional[Union[str, List[str]]] = None,
        query_auth_token_enabled: Optional[bool] = None,
        grpc_options: Optional[Dict[str, Any]] = None,
        http_options: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        if len(kwargs) > 0:
            raise ValueError(f"Unrecognized options: {list(kwargs.keys())}.")

        if name is not None and not isinstance(name, str):
            raise TypeError("'name' must be a string.")

        if image_uri is not None and not isinstance(image_uri, str):
            raise TypeError("'image_uri' must be a string.")

        if compute_config is not None and not isinstance(compute_config, (str, dict)):
            raise TypeError("'compute_config' must be a string or dictionary.")

        if working_dir is not None and not isinstance(working_dir, str):
            raise TypeError("'working_dir' must be a string.")

        if excludes is not None and (
            not isinstance(excludes, list)
            or not all(isinstance(e, str) for e in excludes)
        ):
            raise TypeError("'excludes' must be a list of strings.")

        if not isinstance(applications, list):
            raise TypeError("'applications' must be a list.")
        elif len(applications) == 0:
            raise ValueError("'applications' cannot be empty.")

        if query_auth_token_enabled is not None and not isinstance(
            query_auth_token_enabled, bool
        ):
            raise TypeError("'query_auth_token_enabled' must be a boolean.")

        self._validate_grpc_options(grpc_options)
        self._validate_http_options(http_options)

        self._name = name
        self._image_uri = image_uri
        self._compute_config = compute_config
        self._query_auth_token_enabled = (
            query_auth_token_enabled if query_auth_token_enabled is not None else True
        )
        self._http_options = http_options
        self._grpc_options = grpc_options
        self._applications = self._override_application_runtime_envs(
            applications=applications,
            working_dir=working_dir,
            excludes=excludes,
            requirements=requirements,
        )

        self._validate_import_paths(self._applications)
        if self._image_uri:
            self._validate_image_uri(self._image_uri)

    def _validate_import_paths(self, applications: List[Dict[str, Any]]):
        for app in applications:
            import_path = app.get("import_path", None)
            if not import_path:
                raise ValueError("Every application must specify an import path.")

            if not isinstance(import_path, str):
                raise TypeError(f"'import_path' must be a string, got: {import_path}")

            if (
                import_path.count(":") != 1
                or import_path.rfind(":") in {0, len(import_path) - 1}
                or import_path.rfind(".") in {0, len(import_path) - 1}
            ):
                raise ValueError(
                    f"'import_path' must be of the form: 'module.optional_submodule:app', but got: '{import_path}'."
                )

    def _validate_image_uri(self, image_uri: str):
        matches = IMAGE_URI_PATTERN.match(image_uri)
        if not matches:
            raise ValueError(
                f"Invalid image URI: '{image_uri}'. Must be in the format: [registry_host/]user_name/repository[:tag][@digest]."
            )

    def _validate_grpc_options(self, grpc_options: Optional[Dict[str, Any]]):
        """Validate the `grpc_options` field.

        This will be passed through as part of the Ray Serve config, but some fields are
        disallowed (not valid when deploying Anyscale services).
        """
        if grpc_options is None:
            return
        elif not isinstance(grpc_options, dict):
            raise TypeError("'grpc_options' must be a dict.")

        banned_options = {
            "port",
        }
        banned_options_passed = {o for o in banned_options if o in grpc_options}
        if len(banned_options_passed) > 0:
            raise ValueError(
                "The following provided 'grpc_options' are not permitted "
                f"in Anyscale: {banned_options_passed}."
            )

    def _validate_http_options(self, http_options: Optional[Dict[str, Any]]):
        """Validate the `http_options` field.

        This will be passed through as part of the Ray Serve config, but some fields are
        disallowed (not valid when deploying Anyscale services).
        """
        if http_options is None:
            return
        elif not isinstance(http_options, dict):
            raise TypeError("'http_options' must be a dict.")

        banned_options = {"host", "port", "root_path"}
        banned_options_passed = {o for o in banned_options if o in http_options}
        if len(banned_options_passed) > 0:
            raise ValueError(
                "The following provided 'http_options' are not permitted "
                f"in Anyscale: {banned_options_passed}."
            )

    def _override_application_runtime_envs(
        self,
        applications: List[Dict[str, Any]],
        *,
        working_dir: Optional[str],
        excludes: Optional[List[str]],
        requirements: Union[None, str, List[str]],
    ) -> List[Dict[str, Any]]:
        """Override the runtime_env field of the provided applications.

        Fields that are modified:
            - 'working_dir' is overwritten with the passed working_dir.
            - 'pip' is overwritten with the passed requirements.
            - 'excludes' is extended with the passed excludes list.
        """
        applications = copy.deepcopy(applications)
        for application in applications:
            runtime_env = application.get("runtime_env", {})
            if working_dir is not None:
                runtime_env["working_dir"] = working_dir

            if excludes is not None:
                # Extend the list of excludes rather than overwriting it.
                runtime_env["excludes"] = runtime_env.get("excludes", []) + excludes

            if requirements is not None:
                runtime_env["pip"] = requirements

            if runtime_env:
                application["runtime_env"] = runtime_env

        return applications

    def to_dict(self) -> Dict[str, Any]:
        d = {
            "name": self._name,
            "image_uri": self._image_uri,
            "compute_config": self._compute_config,
            "applications": self._applications,
            "query_auth_token_enabled": self._query_auth_token_enabled,
        }
        if self._grpc_options is not None:
            d["grpc_options"] = self._grpc_options
        if self._http_options is not None:
            d["http_options"] = self._http_options

        return d

    @classmethod
    def from_dict(cls, d: Dict[str, Any]):
        return cls(**d)

    @classmethod
    def from_yaml(cls, path: str):
        with open(path) as f:
            return cls.from_dict(yaml.safe_load(f))

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def image_uri(self) -> Optional[str]:
        return self._image_uri

    @property
    def compute_config(self) -> Optional[Union[str, Dict]]:
        return self._compute_config

    @property
    def applications(self) -> List[Dict[str, Any]]:
        return self._applications

    @property
    def query_auth_token_enabled(self) -> bool:
        return self._query_auth_token_enabled

    @property
    def grpc_options(self) -> Optional[Dict[str, Any]]:
        return self._grpc_options

    @property
    def http_options(self) -> Optional[Dict[str, Any]]:
        return self._http_options

    def options(  # noqa: PLR0913
        self,
        applications: Optional[List[Dict[str, Any]]] = None,
        name: Optional[str] = None,
        image_uri: Optional[str] = None,
        compute_config: Optional[str] = None,
        working_dir: Optional[str] = None,
        excludes: Optional[List[str]] = None,
        requirements: Optional[Union[str, List[str]]] = None,
        query_auth_token_enabled: Optional[bool] = None,
        grpc_options: Optional[Dict[str, Any]] = None,
        http_options: Optional[Dict[str, Any]] = None,
    ) -> "ServiceConfig":
        return ServiceConfig(
            applications=applications or self.applications,
            name=name or self.name,
            image_uri=image_uri or self.image_uri,
            compute_config=compute_config or self.compute_config,
            working_dir=working_dir,
            excludes=excludes,
            requirements=requirements,
            query_auth_token_enabled=query_auth_token_enabled
            if query_auth_token_enabled is not None
            else self.query_auth_token_enabled,
            grpc_options=grpc_options or self.grpc_options,
            http_options=http_options or self.http_options,
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ServiceConfig):
            return all(
                [
                    self.name == other.name,
                    self.image_uri == other.image_uri,
                    self.compute_config == other.compute_config,
                    self.applications == other.applications,
                    self.query_auth_token_enabled == other.query_auth_token_enabled,
                    self.grpc_options == other.grpc_options,
                    self.http_options == other.http_options,
                ]
            )

        return False

    def __str__(self) -> str:
        return str(self.to_dict())


class ServiceState(str, Enum):
    UNKNOWN = "UNKNOWN"
    STARTING = "STARTING"
    RUNNING = "RUNNING"
    # TODO(edoakes): UPDATING comes up while rolling out and rolling back.
    # This is very unexpected from a customer's point of view, we should fix it.
    UPDATING = "UPDATING"
    ROLLING_OUT = "ROLLING_OUT"
    ROLLING_BACK = "ROLLING_BACK"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"
    UNHEALTHY = "UNHEALTHY"
    SYSTEM_FAILURE = "SYSTEM_FAILURE"

    def __str__(self):
        return self.name


# TODO(edoakes): we should have a corresponding ServiceVersionState.
class ServiceVersionStatus:
    def __init__(
        self, *, name: str, weight: int, config: Union[ServiceConfig, Dict],
    ):
        self._name: str = name
        self._weight: int = weight

        if isinstance(config, Dict):
            config = ServiceConfig(**config)

        self._config: ServiceConfig = config

    @property
    def name(self) -> str:
        return self._name

    @property
    def weight(self) -> int:
        return self._weight

    @property
    def config(self) -> ServiceConfig:
        return self._config

    def to_dict(self, *, exclude_details: bool = False) -> Dict[str, Any]:
        d = {
            "name": self._name,
            "weight": self._weight,
        }
        if not exclude_details:
            d["config"] = self._config.to_dict()

        return d

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ServiceVersionStatus):
            return all(
                [
                    self.name == other.name,
                    self.weight == other.weight,
                    self.config == other.config,
                ]
            )

        return False

    def __repr__(self) -> str:
        return f"ServiceVersionStatus(name='{self._name}', weight='{self._weight}')"

    def __str__(self) -> str:
        return str(self.to_dict())


class ServiceStatus:
    def __init__(
        self,
        *,
        service_id: str,
        name: str,
        state: ServiceState,
        query_url: str,
        query_auth_token: Optional[str] = None,
        primary_version: Optional[Union[ServiceVersionStatus, Dict]] = None,
        canary_version: Optional[Union[ServiceVersionStatus, Dict]] = None,
    ):
        self._service_id: str = service_id
        self._name: str = name
        self._state: ServiceState = ServiceState(state)
        self._query_url: str = query_url
        self._query_auth_token: Optional[str] = query_auth_token

        if isinstance(primary_version, dict):
            primary_version = ServiceVersionStatus(**primary_version)
        self._primary_version: Optional[ServiceVersionStatus] = primary_version

        if isinstance(canary_version, dict):
            canary_version = ServiceVersionStatus(**canary_version)
        self._canary_version: Optional[ServiceVersionStatus] = canary_version

    @property
    def service_id(self) -> str:
        return self._service_id

    @property
    def name(self) -> str:
        return self._name

    @property
    def state(self) -> ServiceState:
        return self._state

    @property
    def query_url(self) -> str:
        return self._query_url

    @property
    def query_auth_token(self) -> Optional[str]:
        return self._query_auth_token

    @property
    def primary_version(self) -> Optional[ServiceVersionStatus]:
        return self._primary_version

    @property
    def canary_version(self) -> Optional[ServiceVersionStatus]:
        return self._canary_version

    def to_dict(self, *, exclude_details: bool = False) -> Dict[str, Any]:
        d: Dict[str, Any] = {
            "service_id": self._service_id,
            "name": self._name,
            "state": str(self._state),
            "query_url": self._query_url,
        }
        if self._query_auth_token is not None:
            d["query_auth_token"] = self._query_auth_token

        if self._primary_version is not None:
            d["primary_version"] = self._primary_version.to_dict(
                exclude_details=exclude_details
            )

        if self._canary_version is not None:
            d["canary_version"] = self._canary_version.to_dict(
                exclude_details=exclude_details
            )

        return d

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ServiceStatus):
            return all(
                [
                    self.service_id == other.service_id,
                    self.name == other.name,
                    self.state == other.state,
                    self.query_url == other.query_url,
                    self.query_auth_token == other.query_auth_token,
                    self.primary_version == other.primary_version,
                    self.canary_version == other.canary_version,
                ]
            )

        return False

    def __repr__(self) -> str:
        return f"ServiceStatus(name='{self._name}')"

    def __str__(self) -> str:
        return str(self.to_dict())



import copy
from typing import Any, Dict, List, Optional, Union

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import (
    ComputeTemplateConfig,
    DecoratedComputeTemplateConfig,
)
from anyscale.sdk.anyscale_client.models import (
    AccessConfig,
    ApplyServiceModel,
    ProductionServiceV2VersionModel,
    ServiceConfig as ExternalAPIServiceConfig,
    ServiceEventCurrentState,
    ServiceModel,
)
from anyscale.service._private.anyscale_client import (
    AnyscaleClientWrapper,
    RealAnyscaleClient,
    WORKSPACE_CLUSTER_NAME_PREFIX,
)
from anyscale.service.models import (
    ServiceConfig,
    ServiceState,
    ServiceStatus,
    ServiceVersionStatus,
)
from anyscale.utils.runtime_env import is_dir_remote_uri, parse_requirements_file
from anyscale.utils.workspace_notification import (
    WorkspaceNotification,
    WorkspaceNotificationAction,
)


logger = BlockLogger()


class ServiceSDK:
    def __init__(self, *, client: Optional[AnyscaleClientWrapper] = None):
        self._client = client or RealAnyscaleClient()

    def _override_and_load_requirements_files(
        self, config: ServiceConfig, *, autopopulate_in_workspace: bool = True
    ) -> ServiceConfig:
        """Replaces any pip fields in runtime_envs with their parsed file contents.

        If autopopulate_from_workspace is passed and this code is running inside a
        workspace, any applications that do not have a "pip" or "conda" field
        specified in the runtime_env will have it autopopulated with the
        workspace-tracked dependencies.
        """
        new_applications = copy.deepcopy(config.applications)

        local_path_to_parsed_requirements: Dict[str, List[str]] = {}

        def _load_requirements_file_memoized(
            target: Union[str, List[str]]
        ) -> List[str]:
            if isinstance(target, list):
                return target
            elif target in local_path_to_parsed_requirements:
                return local_path_to_parsed_requirements[target]
            elif isinstance(target, str):
                parsed_requirements = parse_requirements_file(target)
                if parsed_requirements is None:
                    raise FileNotFoundError(
                        f"Requirements file {target} does not exist."
                    )
                local_path_to_parsed_requirements[target] = parsed_requirements
                return parsed_requirements
            else:
                raise TypeError("pip field in runtime_env must be a list or string.")

            return parsed_requirements

        for application in new_applications:
            runtime_env = application.get("runtime_env", {})
            if "pip" in runtime_env:
                # Allow `None` to be used as a signal not to override.
                if runtime_env["pip"] is not None:
                    # If the user passed a "pip" field, load from the file if necessary.
                    runtime_env["pip"] = _load_requirements_file_memoized(
                        runtime_env["pip"]
                    )
            elif (
                "conda" not in runtime_env
                and autopopulate_in_workspace
                and self._client.inside_workspace()
            ):
                # If no "pip" or "conda" field was specified, try filling it in from
                # the workspace-tracked requirements file.
                workspace_requirements_path = (
                    self._client.get_workspace_requirements_path()
                )
                if workspace_requirements_path is not None:
                    logger.info("Including workspace-managed pip dependencies.")
                    runtime_env["pip"] = _load_requirements_file_memoized(
                        self._client.get_workspace_requirements_path()
                    )

        return config.options(applications=new_applications)

    def _override_and_upload_local_dirs(
        self,
        config: ServiceConfig,
        *,
        autopopulate_in_workspace: bool = True,
        cloud_id: Optional[str] = None,
    ) -> ServiceConfig:
        """Returns a copy of the config with all local dirs converted to remote URIs.

        Local dirs can be specified in the working_dir or py_modules fields of the runtime_env.

        Each unique local directory across these fields will be uploaded once to cloud storage,
        then all occurrences of it in the config will be replaced with the corresponding remote URI.
        """
        new_applications = copy.deepcopy(config.applications)

        local_path_to_uri: Dict[str, str] = {}

        def _upload_dir_memoized(target: str, *, excludes: Optional[List[str]]) -> str:
            if is_dir_remote_uri(target):
                return target
            if target in local_path_to_uri:
                return local_path_to_uri[target]

            logger.info(f"Uploading local dir '{target}' to cloud storage.")
            uri = self._client.upload_local_dir_to_cloud_storage(
                target, cloud_id=cloud_id, excludes=excludes,
            )
            local_path_to_uri[target] = uri
            return uri

        for application in new_applications:
            runtime_env = application.get("runtime_env", {})
            excludes = runtime_env.pop("excludes", None)
            if "working_dir" in runtime_env:
                # Allow `None` to be used as a signal not to override.
                if runtime_env["working_dir"] is not None:
                    runtime_env["working_dir"] = _upload_dir_memoized(
                        runtime_env["working_dir"], excludes=excludes
                    )
            elif autopopulate_in_workspace and self._client.inside_workspace():
                runtime_env["working_dir"] = _upload_dir_memoized(
                    ".", excludes=excludes,
                )

            py_modules = runtime_env.get("py_modules", None)
            if py_modules is not None:
                new_py_modules = []
                for py_module in py_modules:
                    new_py_modules.append(
                        _upload_dir_memoized(py_module, excludes=excludes)
                    )

                application["runtime_env"]["py_modules"] = new_py_modules

            if runtime_env:
                application["runtime_env"] = runtime_env

        return config.options(applications=new_applications)

    def _resolve_compute_config_to_id(
        self, compute_config: Optional[Union[str, Dict]]
    ) -> str:
        if compute_config is None:
            compute_config_id = self._client.get_compute_config_id()
            assert compute_config_id is not None
        elif isinstance(compute_config, str):
            compute_config_id = self._client.get_compute_config_id(
                compute_config_name=compute_config,
            )
            if compute_config_id is None:
                raise ValueError(
                    f"The compute config '{compute_config}' does not exist."
                )
        else:
            try:
                compute_config_template = ComputeTemplateConfig(**compute_config)
            except (ValueError, TypeError) as e:
                raise RuntimeError(f"Invalid compute config: {e}.")

            compute_config_id = self._client.create_anonymous_compute_config(
                compute_config_template
            )

        return compute_config_id

    def _get_default_name(self) -> Optional[str]:
        """Get a default name for the service.

        A default is currently only generated when running inside a workspace
        (from the workspace cluster name), so this function errors if called outside
        a workspace.
        """
        if not self._client.inside_workspace():
            raise ValueError(
                "A service name must be provided when running outside of a workspace."
            )

        name = self._client.get_current_workspace_cluster().name
        # Defensively default to the workspace cluster name as-is if it doesn't
        # start with the expected prefix.
        if name.startswith(WORKSPACE_CLUSTER_NAME_PREFIX):
            name = name[len(WORKSPACE_CLUSTER_NAME_PREFIX) :]

        logger.info(f"No name was specified, using default: '{name}'.")
        return name

    def deploy(  # noqa: PLR0912
        self,
        config: ServiceConfig,
        *,
        in_place: bool = False,
        canary_percent: Optional[int] = None,
        max_surge_percent: Optional[int] = None,
    ):
        if not isinstance(in_place, bool):
            raise TypeError("in_place must be a bool.")

        if canary_percent is not None:
            if not isinstance(canary_percent, int):
                raise TypeError("canary_percent must be an int.")
            if canary_percent < 0 or canary_percent > 100:
                raise ValueError("canary_percent must be between 0 and 100.")

        if max_surge_percent is not None:
            if not isinstance(max_surge_percent, int):
                raise TypeError("max_surge_percent must be an int.")

            if max_surge_percent < 0 or max_surge_percent > 100:
                raise ValueError("max_surge_percent must be between 0 and 100.")

        name = config.name or self._get_default_name()

        if config.image_uri is None:
            build_id = self._client.get_default_build_id()
        else:
            build_id = self._client.get_cluster_env_build_id(
                image_uri=config.image_uri,
            )

        compute_config_id = self._resolve_compute_config_to_id(config.compute_config)

        # If a compute config was specified, we need to make sure to used the correct
        # cloud_id when uploading local directories.
        cloud_id = self._client.get_cloud_id(
            compute_config_id=None
            if config.compute_config is None
            else compute_config_id
        )
        config = self._override_and_upload_local_dirs(config, cloud_id=cloud_id)
        config = self._override_and_load_requirements_files(config)

        ray_serve_config: Dict[str, Any] = {"applications": config.applications}
        if config.http_options:
            ray_serve_config["http_options"] = config.http_options
        if config.grpc_options:
            ray_serve_config["grpc_options"] = config.grpc_options
        service: ServiceModel = self._client.rollout_service(
            ApplyServiceModel(
                name=name,
                project_id=self._client.get_project_id(),
                ray_serve_config=ray_serve_config,
                build_id=build_id,
                compute_config_id=compute_config_id,
                canary_percent=canary_percent,
                max_surge_percent=max_surge_percent,
                rollout_strategy="IN_PLACE" if in_place else "ROLLOUT",
                config=ExternalAPIServiceConfig(
                    access=AccessConfig(
                        use_bearer_token=config.query_auth_token_enabled
                    ),
                ),
            )
        )

        canary_percent_info = (
            ""
            if canary_percent is None
            else f" (target canary percent: {canary_percent})"
        )
        message = f"Service '{name}' deployed{canary_percent_info}."
        logger.info(message)
        if self._client.inside_workspace():
            self._client.send_workspace_notification(
                WorkspaceNotification(
                    body=message,
                    action=WorkspaceNotificationAction(
                        type="navigate-service", title="View Service", value=service.id,
                    ),
                ),
            )
        else:
            # NOTE(edoakes): this link does not currently work for AIOA clouds.
            logger.info(
                f"View the service in the UI: {self._client.get_service_ui_url(service.id)}"
            )

        logger.info(
            "Query the service once it's running using the following curl command:"
        )
        auth_token_header = (
            ""
            if service.auth_token is None
            else f"-H 'Authorization: Bearer {service.auth_token}' "
        )
        logger.info(f"curl {auth_token_header}{service.base_url}/")

    def rollback(
        self, name: Optional[str] = None, *, max_surge_percent: Optional[int] = None,
    ):
        if name is None:
            name = self._get_default_name()

        model: Optional[ServiceModel] = self._client.get_service(name)
        if model is None:
            raise RuntimeError(f"Service with name '{name}' was not found.")

        return self._client.rollback_service(
            model.id, max_surge_percent=max_surge_percent,
        )

    def terminate(self, name: Optional[str] = None):
        if name is None:
            name = self._get_default_name()

        model: Optional[ServiceModel] = self._client.get_service(name)
        if model is None:
            raise RuntimeError(f"Service with name '{name}' was not found.")

        return self._client.terminate_service(model.id,)

    def _strip_none_values_from_dict(self, d: Dict) -> Dict:
        """Return a copy of the dictionary without any keys whose values are None.

        Recursively calls into any dictionary values.
        """
        result = {}
        for k, v in d.items():
            if isinstance(v, dict):
                result[k] = self._strip_none_values_from_dict(v)
            elif v is not None:
                result[k] = v

        return result

    def _get_compute_config_for_status(
        self, compute_config_id: str,
    ) -> Union[str, Dict]:
        """Get the compute config to be displayed in the status for this ID.

        If the compute config refers to an anonymous compute config, its config
        will be returned. Else the name of the compute config will be returned.
        """
        compute_config = self._client.get_compute_config(compute_config_id)
        if compute_config is None:
            raise RuntimeError(
                f"Failed to get compute config for ID {compute_config_id}."
            )

        if not compute_config.anonymous:
            return compute_config.name

        config: DecoratedComputeTemplateConfig = compute_config.config
        config_dict = {
            # Strip some internal fields that we don't want to expose to the user.
            # TODO(edoakes): we should improve the compute config model and avoid this.
            k: v
            for k, v in config.to_dict().items()
            if k
            not in {
                "cloud",
                "cloud_id",
                "maximum_uptime_minutes",
                "idle_termination_minutes",
            }
        }
        return self._strip_none_values_from_dict(config_dict)

    def _service_version_model_to_status(
        self,
        model: ProductionServiceV2VersionModel,
        *,
        service_name: str,
        query_auth_token_enabled: bool,
    ) -> ServiceVersionStatus:
        cluster_env_name = self._client.get_cluster_env_name(model.build_id)
        if cluster_env_name is None:
            raise RuntimeError(
                f"Failed to get cluster env name for ID {model.build_id}."
            )

        return ServiceVersionStatus(
            name=model.version,
            # NOTE(edoakes): there is also a "current_weight" field but it does not match the UI.
            weight=model.weight,
            config=ServiceConfig(
                name=service_name,
                applications=model.ray_serve_config["applications"],
                image_uri=cluster_env_name,
                compute_config=self._get_compute_config_for_status(
                    model.compute_config_id
                ),
                query_auth_token_enabled=query_auth_token_enabled,
                http_options=model.ray_serve_config.get("http_options", None),
                grpc_options=model.ray_serve_config.get("grpc_options", None),
            ),
        )

    def _service_model_to_status(self, model: ServiceModel) -> ServiceStatus:
        # TODO(edoakes): for some reason the primary_version is populated
        # when the service is terminated. This should be fixed in the backend.
        is_terminated = model.current_state == ServiceEventCurrentState.TERMINATED

        # TODO(edoakes): this is currently only exposed at the service level in the API,
        # which means that the per-version `query_auth_token_enabled` field will lie if
        # it's changed.
        query_auth_token_enabled = model.auth_token is not None

        primary_version = None
        if not is_terminated and model.primary_version is not None:
            primary_version = self._service_version_model_to_status(
                model.primary_version,
                service_name=model.name,
                query_auth_token_enabled=query_auth_token_enabled,
            )

        canary_version = None
        if not is_terminated and model.canary_version is not None:
            canary_version = self._service_version_model_to_status(
                model.canary_version,
                service_name=model.name,
                query_auth_token_enabled=query_auth_token_enabled,
            )

        # If we add a new state to the backend, old clients may not recognize it.
        # Rather than erroring out and causing old code to crash, return UNKNOWN.
        try:
            state = ServiceState(model.current_state)
        except ValueError:
            state = ServiceState.UNKNOWN
            logger.warning(
                f"Got unrecognized state: '{model.current_state}'. "
                "You likely need to update the 'anyscale' package. "
                "If you still see this message after upgrading, contact Anyscale support."
            )

        return ServiceStatus(
            service_id=model.id,
            name=model.name,
            state=state,
            query_url=model.base_url,
            query_auth_token=model.auth_token,
            primary_version=primary_version,
            canary_version=canary_version,
        )

    def status(self, name: Optional[str] = None) -> ServiceStatus:
        if name is None:
            name = self._get_default_name()

        model: Optional[ServiceModel] = self._client.get_service(name)
        if model is None:
            raise RuntimeError(f"Service with name '{name}' was not found.")

        return self._service_model_to_status(model)



from dataclasses import dataclass
import os
import re
from typing import Any, Dict, List, Optional

import pytest

from anyscale.service.models import (
    ServiceConfig,
    ServiceState,
    ServiceStatus,
    ServiceVersionStatus,
)


@dataclass
class ConfigFile:
    name: str
    expected_config: Optional[ServiceConfig] = None
    expected_error: Optional[str] = None

    def get_path(self) -> str:
        return os.path.join(os.path.dirname(__file__), "test_config_files", self.name)


TEST_CONFIG_FILES = [
    ConfigFile(
        "minimal.yaml",
        expected_config=ServiceConfig(applications=[{"import_path": "main:app"}]),
    ),
    ConfigFile(
        "full.yaml",
        expected_config=ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config="test-compute-config",
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0,},
            grpc_options={"grpc_servicer_functions": ["hello.world"],},
        ),
    ),
    ConfigFile(
        "points_to_requirements_file.yaml",
        expected_config=ServiceConfig(
            applications=[{"import_path": "main:app"}],
            requirements="some_requirements_file.txt",
        ),
    ),
    ConfigFile(
        "unrecognized_option.yaml",
        expected_error=re.escape("Unrecognized options: ['bad_option']."),
    ),
    ConfigFile(
        "multiple_applications.yaml",
        expected_config=ServiceConfig(
            applications=[
                {
                    "name": "app1",
                    "import_path": "main:app1",
                    "runtime_env": {"env_vars": {"abc": "def"}},
                },
                {
                    "name": "app2",
                    "import_path": "main:app",
                    "arguments": {"abc": "def", "nested": {"key": "val"}},
                },
            ],
        ),
    ),
]


class TestServiceConfig:
    def test_empty_applications(self):
        with pytest.raises(ValueError, match="'applications' cannot be empty."):
            ServiceConfig(applications=[])

    def test_bad_applications_type(self):
        with pytest.raises(TypeError, match="'applications' must be a list."):
            ServiceConfig(applications={})

    def test_invalid_import_paths(self):
        with pytest.raises(TypeError, match="'import_path' must be a string"):
            ServiceConfig(applications=[{"import_path": 1}])

        with pytest.raises(ValueError, match="'import_path' must be"):
            ServiceConfig(applications=[{"import_path": "hello"}])

        with pytest.raises(ValueError, match="'import_path' must be"):
            ServiceConfig(applications=[{"import_path": "hello."}])

        with pytest.raises(ValueError, match="'import_path' must be"):
            ServiceConfig(applications=[{"import_path": "hello:"}])

        with pytest.raises(ValueError, match="'import_path' must be"):
            ServiceConfig(applications=[{"import_path": ":hello"}])

        with pytest.raises(ValueError, match="'import_path' must be"):
            ServiceConfig(applications=[{"import_path": ".hello"}])

    def test_applications(self):
        applications = [
            {"import_path": "main:app"},
            {"import_path": "main:app", "name": "app2"},
        ]
        config = ServiceConfig(applications=applications)
        assert config.applications == applications

    def test_name(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert config.name is None

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="my-custom-name"
        )
        assert config.name == "my-custom-name"

        with pytest.raises(TypeError, match="'name' must be a string"):
            ServiceConfig(applications=[{"import_path": "main:app"}], name=123)

    def test_image_uri(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert config.image_uri is None

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            image_uri="user/my-custom-image:1",
        )
        assert config.image_uri == "user/my-custom-image:1"

        with pytest.raises(TypeError, match="'image_uri' must be a string"):
            ServiceConfig(applications=[{"import_path": "main:app"}], image_uri=123)

    def test_compute_config(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert config.compute_config is None

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            compute_config="my-custom-compute_config",
        )
        assert config.compute_config == "my-custom-compute_config"

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            compute_config={"test": "inlined"},
        )
        assert config.compute_config == {"test": "inlined"}

        with pytest.raises(
            TypeError, match="'compute_config' must be a string or dictionary."
        ):
            ServiceConfig(
                applications=[{"import_path": "main:app"}], compute_config=123
            )

    def test_query_auth_token_enabled(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}],)
        assert config.query_auth_token_enabled is True

        config = ServiceConfig(applications=[{"import_path": "main:app"}],)
        assert config.query_auth_token_enabled is True

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], query_auth_token_enabled=False,
        )
        assert config.query_auth_token_enabled is False
        assert config.options().query_auth_token_enabled is False
        assert (
            config.options(query_auth_token_enabled=True).query_auth_token_enabled
            is True
        )

        with pytest.raises(
            TypeError, match="'query_auth_token_enabled' must be a boolean."
        ):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                query_auth_token_enabled="foobar",
            )

    def test_http_options(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}],)
        assert config.http_options is None

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            http_options={"request_timeout_s": 10.0},
        )
        assert config.http_options == {"request_timeout_s": 10.0}

        with pytest.raises(TypeError, match="'http_options' must be a dict."):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                http_options=["request_timeout_s"],
            )

        with pytest.raises(
            ValueError,
            match="The following provided 'http_options' are not permitted in Anyscale: {'host'}.",
        ):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                http_options={"request_timeout_s": 10.0, "host": "0.0.0.0"},
            )

    def test_grpc_options(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}],)
        assert config.grpc_options is None

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
        )
        assert config.grpc_options == {"grpc_servicer_functions": ["hello.world"]}

        with pytest.raises(TypeError, match="'grpc_options' must be a dict."):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                grpc_options=["grpc_servicer_functions"],
            )

        with pytest.raises(
            ValueError,
            match="The following provided 'http_options' are not permitted in Anyscale: {'port'}.",
        ):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                http_options={"grpc_servicer_functions": ["hello.world"], "port": 9001},
            )

    def test_options(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}])

        options = {
            "name": "test-name",
            "image_uri": "docker.io/libaray/test-image:latest",
            "compute_config": "test-compute-config",
            "requirements": ["pip-install-test"],
            "working_dir": ".",
            "excludes": ["some-path"],
            "query_auth_token_enabled": False,
            "http_options": {"request_timeout_s": 10.0},
            "grpc_options": {"grpc_servicer_functions": ["hello.world"]},
        }

        # Test setting fields one at a time.
        for option, val in options.items():
            assert config.options(**{option: val}) == ServiceConfig(
                applications=[{"import_path": "main:app"}], **{option: val}
            )

            assert config.options(**{option: val}) == ServiceConfig(
                applications=[{"import_path": "main:app"}], **{option: val}
            )

        # Test setting fields all at once.
        assert config.options(**options) == ServiceConfig(
            applications=[{"import_path": "main:app"}], **options
        )

    def test_working_dir_excludes_bad_types(self):
        with pytest.raises(TypeError, match="'working_dir' must be a string."):
            ServiceConfig(
                applications=[{"import_path": "main:app"}], working_dir=1,
            )

        with pytest.raises(TypeError, match="'excludes' must be a list of strings."):
            ServiceConfig(
                applications=[{"import_path": "main:app"}], excludes="hi",
            )

        with pytest.raises(TypeError, match="'excludes' must be a list of strings."):
            ServiceConfig(
                applications=[{"import_path": "main:app"}], excludes=["hi", 1],
            )

    @pytest.mark.parametrize(
        ("working_dir", "excludes"),
        [
            (None, None),
            (".", ["path1/", "path2"]),
            ("s3://path.zip", ["path1/", "path2"]),
        ],
    )
    def test_override_working_dir_excludes(
        self, working_dir: Optional[str], excludes: Optional[List[str]]
    ):
        applications: List[Dict[str, Any]] = [
            {"name": "no_runtime_env", "import_path": "main:app"},
            {
                "name": "empty_runtime_env",
                "import_path": "main:app",
                "runtime_env": {},
            },
            {
                "name": "has_runtime_env",
                "import_path": "main:app",
                "runtime_env": {
                    "excludes": ["something-else"],
                    "working_dir": "s3://somewhere.zip",
                    "env_vars": {"abc": "123"},
                },
            },
        ]

        config = ServiceConfig(
            applications=applications, working_dir=working_dir, excludes=excludes
        )

        if working_dir is None:
            assert excludes is None
            assert config.applications == applications
        else:
            assert isinstance(excludes, list)
            assert len(config.applications) == 3
            assert config.applications[0] == {
                "name": "no_runtime_env",
                "import_path": "main:app",
                "runtime_env": {"working_dir": working_dir, "excludes": excludes},
            }

            assert config.applications[1] == {
                "name": "empty_runtime_env",
                "import_path": "main:app",
                "runtime_env": {"working_dir": working_dir, "excludes": excludes},
            }
            assert config.applications[2] == {
                "name": "has_runtime_env",
                "import_path": "main:app",
                "runtime_env": {
                    "working_dir": working_dir,
                    # Existing `excludes` field should be extended.
                    "excludes": ["something-else", *excludes],
                    "env_vars": {"abc": "123"},
                },
            }

    @pytest.mark.parametrize("config_file", TEST_CONFIG_FILES)
    def test_from_config_file(self, config_file: ConfigFile):
        if config_file.expected_error is not None:
            with pytest.raises(Exception, match=config_file.expected_error):
                ServiceConfig.from_yaml(config_file.get_path())

            return

        assert config_file.expected_config == ServiceConfig.from_yaml(
            config_file.get_path()
        )

    @pytest.mark.parametrize(
        ("image_uri", "valid"),
        [
            ("docker.io/libaray/ubuntu:latest", True),
            ("ubuntu:latest", False),
            ("python:3.8", False),
            ("myregistry.local:5000/testing/test-image:1.0.0", True),
            ("localhost:5000/myusername/myrepository:latest", True),
            ("valid/withouttag", True),
            ("valid_name/withtag_and_digest:v2@sha213", True),
            ("valid_name/withtag_and_digest@sha213", True),
            ("valid_name/withtag_and_digest:@sha213", False),
            ("http://myregistry.local:5000/testing/test-image:1.0.0", False),
            (
                "us-docker.pkg.dev/anyscale-workspace-templates/workspace-templates/rag-dev-bootcamp-mar-2024:raynightly-py310",
                True,
            ),
        ],
    )
    def test_image_uri_validation(self, image_uri: str, valid: bool):
        if valid:
            _ = ServiceConfig(
                applications=[{"import_path": "main:app"}], image_uri=image_uri
            )
        else:
            with pytest.raises(
                ValueError,
                match=re.escape(
                    f"Invalid image URI: '{image_uri}'. Must be in the format: [registry_host/]user_name/repository[:tag][@digest]."
                ),
            ):
                _ = ServiceConfig(
                    applications=[{"import_path": "main:app"}], image_uri=image_uri
                )


class TestServiceStatus:
    @pytest.mark.parametrize(
        "canary_version",
        [
            None,
            ServiceVersionStatus(
                name="test-canary-version",
                weight=0,
                config=ServiceConfig(
                    applications=[
                        {"import_path": "main:app", "arguments": {"version": "canary"}}
                    ],
                    image_uri="docker.io/library/test-image:latest",
                    compute_config="test-compute-config",
                ),
            ),
        ],
    )
    @pytest.mark.parametrize(
        "primary_version",
        [
            None,
            ServiceVersionStatus(
                name="test-primary-version",
                weight=100,
                config=ServiceConfig(
                    applications=[
                        {"import_path": "main:app", "arguments": {"version": "primary"}}
                    ],
                    image_uri="docker.io/library/test-image:latest",
                    compute_config="test-compute-config",
                ),
            ),
        ],
    )
    @pytest.mark.parametrize("query_auth_token", [None, "test_abc123"])
    @pytest.mark.parametrize("state", list(ServiceState))
    def test_to_dict_and_back(
        self,
        state: ServiceState,
        query_auth_token: Optional[str],
        primary_version: Optional[ServiceVersionStatus],
        canary_version: Optional[ServiceVersionStatus],
    ):
        """Test that all fields can be serialized to and from dictionaries."""
        status = ServiceStatus(
            service_id="test-service-id",
            name="test-service-name",
            state=state,
            query_url="http://test.com/",
            query_auth_token=query_auth_token,
            primary_version=primary_version,
            canary_version=canary_version,
        )

        status_dict = status.to_dict()
        assert ServiceStatus(**status_dict) == status

    def test_unknown_state(self):
        with pytest.raises(
            ValueError, match="'SOME_FAKE_NEWS_STATE' is not a valid ServiceState"
        ):
            ServiceStatus(
                service_id="test-service-id",
                name="test-service-name",
                state="SOME_FAKE_NEWS_STATE",
                query_url="http://test.com/",
                query_auth_token=None,
                primary_version=None,
                canary_version=None,
            )
